# ZAP - Zurich Atmosphere Purge
#
# Copyright (c) 2014-2016 Kurt Soto
# Copyright (c) 2015-2019 Simon Conseil
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import astropy.units as u
import logging
import numpy as np
import os
import scipy.ndimage as ndi
import sys
import warnings
from astropy.io import fits
from astropy.wcs import WCS
from functools import wraps
from multiprocessing import cpu_count, Manager, Process
from scipy.stats import sigmaclip
from sklearn.decomposition import PCA
from time import time
import esorexplugin
from esorexplugin import *
from pkg_resources import get_distribution, DistributionNotFound
__version__ = '2.1.dev workflow' #get_distribution('zap').version

__all__ = ['process', 'SVDoutput', 'nancleanfits', 'contsubfits', 'Zap',
           'SKYSEG', '__version__']

# Limits of the segments in Angstroms. Zap now uses by default only one
# segment, see below for the old values.
SKYSEG = [0, 10000]

# These are the old limits from the original zap. Keeping them for reference,
# and may be useful for further testing.
# [0, 5400, 5850, 6440, 6750, 7200, 7700, 8265, 8602, 8731, 9275, 10000]

# range where the NaD notch filters absorbs significant flux
NOTCH_FILTER_RANGES = {
    'WFM-AO-E': [5755, 6008],
    'WFM-AO-N': [5805, 5966],
    'NFM-AO-N': [5780, 6050],
}

# List of allowed values for cftype (continuum filter)
CFTYPE_OPTIONS = ('median', 'fit', 'none')

# Number of available CPUs
NCPU = cpu_count()

logging.basicConfig(format='[%(levelname)s] %(message)s', level=logging.INFO,
                    stream=sys.stdout)
logger = logging.getLogger(__name__)


# ================= Top Level Functions =================
try:
 def processzap(cubefits, outcubefits='DATACUBE_ZAP.fits', clean=True,
            zlevel='median', cftype='median', cfwidthSVD=300, cfwidthSP=300,
            nevals=[], extSVD=None, skycubefits=None, mask=None,
            interactive=False, ncpu=None, pca_class=None, n_components=None,
            overwrite=False, varcurvefits=None):
    """ Performs the entire ZAP sky subtraction algorithm.
    This is the main ZAP function. It works on an input FITS file and
    optionally writes the product to an output FITS file.
    Parameters
    ----------
    cubefits : str
        Input FITS file, containing a cube with data in the first extension.
    outcubefits : str
        Output FITS file, based on the input one to propagate all header
        information and other extensions. Default to `DATACUBE_ZAP.fits`.
    clean : bool
        If True (default value), the NaN values are cleaned. Spaxels with more
        then 25% of NaN values are removed, the others are replaced with an
        interpolation from the neighbors. The NaN values are reinserted into
        the final datacube. If set to False, any spaxel with a NaN value will
        be ignored.
    zlevel : str
        Method for the zeroth order sky removal: `none`, `sigclip` or `median`
        (default).
    cftype : {'median', 'fit', 'none'}
        Method for the continuum filter.
    cfwidthSVD : int or float
        Window size for the continuum filter, for the SVD computation.
        Default to 300.
    cfwidthSP : int or float
        Window size for the continuum filter used to remove the continuum
        features for calculating the eigenvalues per spectrum. Smaller values
        better trace the sources. An optimal range of is typically
        20 - 50 pixels. Default to 300.
    nevals : list
        Allow to specify the number of eigenspectra used for each segment.
        Provide either a single value that will be used for all of the
        segments, or a list of 11 values that will be used for each of the
        segments.
    extSVD : Zap object
        Can be a ``Zap`` object output from :func:`~zap.SVDoutput`.
        If given, the SVD from this object will be used, otherwise the SVD is
        computed. So this allows to compute the SVD on an other field or with
        different settings.
    skycubefits : str
        Path for the optional output of the sky that is subtracted from the
        cube. This is simply the input cube minus the output cube.
    mask : str
        A 2D fits image to exclude regions that may contaminate the zlevel or
        eigenspectra. This image should be constructed from the datacube itself
        to match the dimensionality. Sky regions should be marked as 0, and
        astronomical sources should be identified with an integer greater than
        or equal to 1. Default to None.
    interactive : bool
        If True, a :class:`~zap.Zap` object containing all information on
        the ZAP process is returned, and can be used to explore the
        eigenspectra and recompute the output (with the
        :meth:`~zap.Zap.reprocess` method). In this case, the output files
        are not saved (`outcubefits` and `skycubefits` are ignored). Default
        to False.
    varcurvefits : str
        Path for the optional output of the explained variance curves.
    """
    logger.info('Running ZAP %s !', __version__)
    t0 = time()
    try:
        bstr = basestring
    except NameError:
        bstr = str
    if not isinstance(cubefits, bstr):
        raise TypeError('The process method only accepts a single datacube '
                        'filename.')
    # check if outcubefits/skycubefits exists before beginning
    if not overwrite:
        def _check_file_exists(filename):
            if filename is not None and os.path.exists(filename):
                raise IOError('Output file "{0}" exists'.format(filename))

        _check_file_exists(outcubefits)
        _check_file_exists(skycubefits)

    if ncpu is not None:
        global NCPU
        NCPU = ncpu

    if extSVD is not None and mask is not None:
        raise ValueError('extSVD and mask parameters are incompatible: if mask'
                         ' must be used, then the SVD has to be recomputed')

    if mask is not None or (extSVD is None and cfwidthSVD != cfwidthSP):
        # Compute the SVD separately, only if a mask is given, or if the
        # cfwidth values differ and extSVD is not given. Otherwise, the SVD
        # will be computed in the _run method, which allows to avoid running
        # twice the zlevel and continuumfilter steps.
        extSVD = SVDoutput(cubefits, clean=clean, zlevel=zlevel,
                           cftype=cftype, cfwidth=cfwidthSVD, mask=mask)

    zobj = Zap(cubefits, pca_class=pca_class, n_components=n_components)
    zobj._run(clean=clean, zlevel=zlevel, cfwidth=cfwidthSP, cftype=cftype,
              nevals=nevals, extSVD=extSVD)

    if interactive:
        # Return the zobj object without saving files
        return zobj

    if skycubefits is not None:
        zobj.writeskycube(skycubefits=skycubefits, overwrite=overwrite)

    if varcurvefits is not None:
        zobj.writevarcurve(varcurvefits=varcurvefits, overwrite=overwrite)

    zobj.mergefits(outcubefits, overwrite=overwrite)
    logger.info('Zapped! (took %.2f sec.)', time() - t0)


 def SVDoutput(cubefits, clean=True, zlevel='median', cftype='median',
              cfwidth=300, mask=None, ncpu=None, pca_class=None,
              n_components=None):
    """Performs the SVD decomposition of a datacube.
    This allows to use the SVD for a different datacube. It used to allow to
    save the SVD to a file but this is no more possible. Instead it returns
    a ``Zap`` which can be given to the :func:`~zap.process` function.
    Parameters
    ----------
    cubefits : str
        Input FITS file, containing a cube with data in the first extension.
    clean : bool
        If True (default value), the NaN values are cleaned. Spaxels with more
        then 25% of NaN values are removed, the others are replaced with an
        interpolation from the neighbors.
    zlevel : str
        Method for the zeroth order sky removal: `none`, `sigclip` or `median`
        (default).
    cftype : {'median', 'fit', 'none'}
        Method for the continuum filter.
    cfwidth : int or float
        Window size for the continuum filter, default to 300.
    mask : str
        Path of a FITS file containing a mask (1 for objects, 0 for sky).
    """
    logger.info('Processing %s to compute the SVD', cubefits)

    if ncpu is not None:
        global NCPU
        NCPU = ncpu

    zobj = Zap(cubefits, pca_class=pca_class, n_components=n_components)
    zobj._prepare(clean=clean, zlevel=zlevel, cftype=cftype,
                  cfwidth=cfwidth, mask=mask)
    zobj._msvd()
    return zobj


 def contsubfits(cubefits, outfits='CONTSUB_CUBE.fits', ncpu=None,
                cftype='median', cfwidth=300, clean_nan=True, zlevel='median',
                overwrite=False):
    """A standalone implementation of the continuum removal."""
    if ncpu is not None:
        global NCPU
        NCPU = ncpu

    zobj = Zap(cubefits)
    zobj._prepare(clean=clean_nan, zlevel=zlevel, cftype=cftype,
                  cfwidth=cfwidth)
    cube = zobj.make_contcube()

    outhead = _newheader(zobj)
    outhdu = fits.PrimaryHDU(data=cube, header=outhead)
    outhdu.writeto(outfits, overwrite=overwrite)
    logger.info('Continuum cube file saved to %s', outfits)


 def nancleanfits(cubefits, outfn='NANCLEAN_CUBE.fits', rejectratio=0.25,
                 boxsz=1, overwrite=False):
    """Interpolates NaN values from the nearest neighbors.
    Parameters
    ----------
    cubefits : str
        Input FITS file, containing a cube with data in the first extension.
    outfn : str
        Output FITS file. Default to ``NANCLEAN_CUBE.fits``.
    rejectratio : float
        Defines a cutoff for the ratio of NAN to total pixels in a spaxel
        before the spaxel is avoided completely. Default to 0.25
    boxsz : int
        Defines the number of pixels around the offending NaN pixel.
        Default to 1, which looks for the 26 nearest neighbors which
        is a 3x3x3 cube.
    """
    with fits.open(cubefits) as hdu:
        hdu[1].data = _nanclean(hdu[1].data, rejectratio=rejectratio,
                                boxsz=boxsz)[0]
        hdu.writeto(outfn, overwrite=overwrite)


 def timeit(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        t0 = time()
        res = func(*args, **kwargs)
        logger.info('%s - Time: %.2f sec.', func.__name__, time() - t0)
        return res
    return wrapped


# ================= Main class =================

 class Zap(object):

    """ Main class to run each of the steps of ZAP.
    Attributes
    ----------
    cleancube : numpy.ndarray
        The final datacube after removing all of the residual features.
    contarray : numpy.ndarray
        A 2D array containing the subtracted continuum per spaxel.
    cube : numpy.ndarray
        The original cube with the zlevel subtraction performed per spaxel.
    laxis : numpy.ndarray
        A 1d array containing the wavelength solution generated from the header
        parameters.
    wcs : astropy.wcs.WCS
        WCS object with the wavelength solution.
    lranges : list
        A list of the wavelength bin limits used in segmenting the sepctrum
        for SVD.
    nancube : numpy.ndarray
        A 3d boolean datacube containing True in voxels where a NaN value was
        replaced with an interpolation.
    nevals : numpy.ndarray
        A 1d array containing the number of eigenvalues used per segment to
        reconstruct the residuals.
    normstack : numpy.ndarray
        A normalized version of the datacube decunstructed into a 2d array.
    pranges : numpy.ndarray
        The pixel indices of the bounding regions for each spectral segment.
    recon : numpy.ndarray
        A 2d array containing the reconstructed emission line residuals.
    run_clean : bool
        Boolean that indicates that the NaN cleaning method was used.
    run_zlevel : bool
        Boolean indicating that the zero level correction was used.
    stack : numpy.ndarray
        The datacube deconstructed into a 2d array for use in the the SVD.
    variancearray : numpy.ndarray
        A list of length nsegments containing variances calculated per spaxel
        used for normalization
    y,x : numpy.ndarray
        The position in the cube of the spaxels that are in the 2d
        deconstructed stack
    zlsky : numpy.ndarray
        A 1d array containing the result of the zero level subtraction
    """

    def __init__(self, cubefits, pca_class=None, n_components=None):
        self.cubefits = cubefits
        self.ins_mode = None

        with fits.open(cubefits) as hdul:
            self.instrument = hdul[0].header.get('INSTRUME')
            if self.instrument == 'MUSE':
                self.ins_mode = hdul[0].header.get('HIERARCH ESO INS MODE')
                self.cube = hdul[1].data
                self.header = hdul[1].header
            elif self.instrument == 'KCWI':
                self.cube = hdul[0].data
                self.header = hdul[0].header
            else:
                raise ValueError('unsupported instrument %s' % self.instrument)

        # Workaround for floating points errors in wcs computation: if cunit is
        # specified, wcslib will convert in meters instead of angstroms, so we
        # remove cunit before creating the wcs object
        header = self.header.copy()
        unit = u.Unit(header.pop('CUNIT3'))
        self.wcs = WCS(header).sub([3])

        # Create Lambda axis
        wlaxis = np.arange(self.cube.shape[0])
        self.laxis = self.wcs.all_pix2world(wlaxis, 0)[0]
        if unit != u.angstrom:
            # Make sure lambda is in angstroms
            self.laxis = (self.laxis * unit).to(u.angstrom).value

        # Change laser region into zeros if AO
        if self.ins_mode in NOTCH_FILTER_RANGES:
            logger.info('Cleaning laser region for AO, mode=%s, limits=%s',
                        self.ins_mode, NOTCH_FILTER_RANGES[self.ins_mode])
            self.notch_limits = self.wcs.all_world2pix(
                NOTCH_FILTER_RANGES[self.ins_mode], 0)[0].astype(int)
            lmin, lmax = self.notch_limits
            self.cube[lmin:lmax + 1] = 0.0
        else:
            self.notch_limits = None

        # NaN Cleaning
        self.run_clean = False
        self.nancube = None
        self._boxsz = 1
        self._rejectratio = 0.25

        # Mask file
        self.maskfile = None

        # zlevel parameters
        self.run_zlevel = False
        self.zlsky = np.zeros_like(self.laxis)

        # Extraction results
        self.stack = None
        self.y = None
        self.x = None

        # Normalization Maps
        self.contarray = None
        self.variancearray = None
        self.normstack = None

        # identify the spectral range of the dataset
        laxmin = min(self.laxis)
        laxmax = max(self.laxis)

        # List of segmentation limits in the optical
        skyseg = np.array(SKYSEG)
        skyseg = skyseg[(skyseg > laxmin) & (skyseg < laxmax)]

        # segment limit in angstroms
        self.lranges = (np.vstack([np.append(laxmin - 10, skyseg),
                                   np.append(skyseg, laxmax + 10)])).T

        # segment limit in pixels
        laxis = self.laxis
        lranges = self.lranges
        pranges = []
        for i in range(len(lranges)):
            paxis = wlaxis[(laxis > lranges[i, 0]) & (laxis <= lranges[i, 1])]
            pranges.append((np.min(paxis), np.max(paxis) + 1))
        self.pranges = np.array(pranges)

        # eigenspace Subset
        if pca_class is not None:
            logger.info('Using %s', pca_class)
            self.pca_class = pca_class
        else:
            self.pca_class = PCA

        # Reconstruction of sky features
        self.n_components = n_components
        self.recon = None
        self.cleancube = None

    @timeit
    def _prepare(self, clean=True, zlevel='median', cftype='median',
                 cfwidth=300, extzlevel=None, mask=None):
        # clean up the nan values
        if clean:
            self._nanclean()

        # if mask is supplied, apply it
        if mask is not None:
            self._applymask(mask)

        # Extract the spectra that we will be working with
        self._extract()

        # remove the median along the spectral axis
        if extzlevel is None:
            if zlevel.lower() != 'none':
                self._zlevel(calctype=zlevel)
        else:
            self._externalzlevel(extzlevel)

        # remove the continuum level - this is multiprocessed to speed it up
        self._continuumfilter(cfwidth=cfwidth, cftype=cftype)

        # normalize the variance in the segments.
        self._normalize_variance()

    def _run(self, clean=True, zlevel='median', cftype='median',
             cfwidth=300, nevals=[], extSVD=None):
        """ Perform all steps to ZAP a datacube:
        - NaN re/masking,
        - deconstruction into "stacks",
        - zerolevel subraction,
        - continuum removal,
        - normalization,
        - singular value decomposition,
        - eigenvector selection,
        - residual reconstruction and subtraction,
        - data cube reconstruction.
        """
        self._prepare(clean=clean, zlevel=zlevel, cftype=cftype,
                      cfwidth=cfwidth, extzlevel=extSVD)

        # do the multiprocessed SVD calculation
        if extSVD is None:
            self._msvd()
        else:
            self.models = extSVD.models

        self.components = [m.components_.copy() for m in self.models]

        # choose some fraction of eigenspectra or some finite number of
        # eigenspectra
        if nevals == []:
            self.optimize()
            self.chooseevals(nevals=self.nevals)
        else:
            self.chooseevals(nevals=nevals)

        # reconstruct the sky residuals using the subset of eigenspace
        self.reconstruct()

        # stuff the new spectra back into the cube
        self.remold()

    def _nanclean(self):
        """
        Detects NaN values in cube and removes them by replacing them with an
        interpolation of the nearest neighbors in the data cube. The positions
        in the cube are retained in nancube for later remasking.
        """
        self.cube, self.nancube = _nanclean(
            self.cube, rejectratio=self._rejectratio, boxsz=self._boxsz)
        self.run_clean = True

    @timeit
    def _extract(self):
        """Deconstruct the datacube into a 2d array.
        Since spatial information is not required, and the linear algebra
        routines require 2d arrays. The operation rejects any spaxel with even
        a single NaN value, since this would cause the linear algebra routines
        to crash.
        Adds the x and y data of these positions into the Zap class
        """
        # make a map of spaxels with NaNs
        badmap = (np.logical_not(np.isfinite(self.cube))).sum(axis=0)
        # get positions of those with no NaNs
        self.y, self.x = np.where(badmap == 0)
        # extract those positions into a 2d array
        self.stack = self.cube[:, self.y, self.x]
        logger.info('Extract to 2D, %d valid spaxels (%d%%)', len(self.x),
                    len(self.x) / np.prod(self.cube.shape[1:]) * 100)

    def _externalzlevel(self, extSVD):
        """Remove the zero level from the extSVD file."""
        logger.debug('Using external zlevel from %s', extSVD)
        if isinstance(extSVD, Zap):
            self.zlsky = np.array(extSVD.zlsky, copy=True)
            self.run_zlevel = extSVD.run_zlevel
        else:
            self.zlsky = fits.getdata(extSVD, 0)
            self.run_zlevel = 'extSVD'
        self.stack -= self.zlsky[:, np.newaxis]

    @timeit
    def _zlevel(self, calctype='median'):
        """
        Removes a 'zero' level from each spectral plane. Spatial information is
        not required, so it operates on the extracted stack.
        Operates on stack, leaving it with this level removed and adds the data
        'zlsky' to the class. zlsky is a spectrum of the zero levels.
        This zero level is currently calculated with a median.
        Experimental operations -
        - exclude top quartile
        - run in an iterative sigma clipped mode
        """
        self.run_zlevel = calctype
        if calctype == 'none':
            logger.info('Skipping zlevel subtraction')
            return

        if calctype == 'median':
            logger.info('Median zlevel subtraction')
            func = _imedian
        elif calctype == 'sigclip':
            logger.info('Iterative Sigma Clipping zlevel subtraction')
            func = _isigclip
        else:
            raise ValueError('Unknow zlevel type, must be none, median, or '
                             'sigclip')

        self.zlsky = np.hstack(parallel_map(func, self.stack, NCPU, axis=0))
        self.stack -= self.zlsky[:, np.newaxis]

    @timeit
    def _continuumfilter(self, cfwidth=300, cftype='median'):
        """A multiprocessed implementation of the continuum removal.
        This process distributes the data to many processes that then
        reassemble the data.  Uses two filters, a small scale (less than the
        line spread function) uniform filter, and a large scale median filter
        to capture the structure of a variety of continuum shapes.
        added to class
        contarray - the removed continuua
        normstack - "normalized" version of the stack with the continuua
            removed
        """
        if cftype not in CFTYPE_OPTIONS:
            raise ValueError("cftype must be median, fit or none, got {}"
                             .format(cftype))
        logger.info('Applying Continuum Filter, cftype=%s', cftype)
        self._cftype = cftype
        self._cfwidth = cfwidth

        # remove continuum features
        if cftype == 'none':
            self.normstack = self.stack.copy()
        else:
            if cftype == 'fit' and self.instrument != 'MUSE':
                warnings.warn('the continuum fit method is currently adapted '
                              'to MUSE and should not be used for other '
                              'instruments', UserWarning)

            self.contarray = _continuumfilter(self.stack, cftype,
                                              cfwidth=cfwidth,
                                              notch_limits=self.notch_limits)
            self.normstack = self.stack - self.contarray

    def _normalize_variance(self):
        """Normalize the variance in the segments."""
        logger.debug('Normalizing variances')
        # self.variancearray = np.std(self.stack, axis=1)
        # self.normstack /= self.variancearray[:, np.newaxis]

        nseg = len(self.pranges)
        self.variancearray = var = np.zeros((nseg, self.stack.shape[1]))
        for i in range(nseg):
            pmin, pmax = self.pranges[i]
            var[i, :] = np.var(self.normstack[pmin:pmax, :], axis=0)
            self.normstack[pmin:pmax, :] /= var[i, :]

    @timeit
    def _msvd(self):
        """Multiprocessed singular value decomposition.
        Takes the normalized, spectral segments and distributes them
        to the individual svd methods.
        """
        logger.info('Calculating SVD on %d segments (%s)', len(self.pranges),
                    self.pranges)
        indices = [x[0] for x in self.pranges[1:]]
        # normstack = self.stack - self.contarray
        Xarr = np.array_split(self.normstack.T, indices, axis=1)

        self.models = []
        for i, x in enumerate(Xarr):
            if self.n_components is not None:
                ncomp = max(x.shape[1] * self.n_components, 60)
                logger.info('Segment %d, computing %d eigenvectors out of %d',
                            i, ncomp, x.shape[1])
            else:
                ncomp = None

            self.models.append(self.pca_class(n_components=ncomp).fit(x))

    def chooseevals(self, nevals=[]):
        """Choose the number of eigenspectra/evals to use for reconstruction.
        User supplies the number of eigen spectra to be used (neval) or the
        percentage of the eigenspectra that were calculated (peval) from each
        spectral segment to be used.
        The user can either provide a single value to be used for all segments,
        or provide an array that defines neval or peval per segment.
        """
        nranges = len(self.pranges)
        nevals = np.atleast_1d(nevals)

        # deal with an input list
        if len(nevals) > 1:
            if len(nevals) != nranges:
                nevals = np.array([nevals[0]])
                logger.info('Chosen eigenspectra array does not correspond to '
                            'number of segments')
            else:
                logger.info('Choosing %s eigenspectra for segments', nevals)

        if len(nevals) == 1:
            logger.info('Choosing %s eigenspectra for all segments', nevals)
            nevals = np.zeros(nranges, dtype=int) + nevals

        if nevals.ndim == 1:
            start = np.zeros(nranges, dtype=int)
            end = nevals
        else:
            start, end = nevals.T

        self.nevals = nevals
        for i, model in enumerate(self.models):
            model.components_ = self.components[i][start[i]:end[i]]

    @timeit
    def reconstruct(self):
        """Reconstruct the residuals from a given set of eigenspectra and
        eigenvalues
        """
        logger.info('Reconstructing Sky Residuals')
        indices = [x[0] for x in self.pranges[1:]]
        # normstack = self.stack - self.contarray
        Xarr = np.array_split(self.normstack.T, indices, axis=1)
        Xnew = [model.transform(x)
                for model, x in zip(self.models, Xarr)]
        Xinv = [model.inverse_transform(x)
                for model, x in zip(self.models, Xnew)]
        self.recon = np.concatenate([x.T * self.variancearray[i, :]
                                     for i, x in enumerate(Xinv)])
        # self.recon = np.concatenate([x.T for x in Xinv])
        # self.recon *= self.variancearray[:, np.newaxis]

    def make_cube_from_stack(self, stack, with_nans=False):
        """Stuff the stack back into a cube."""
        cube = self.cube.copy()
        cube[:, self.y, self.x] = stack
        if with_nans:
            cube[self.nancube] = np.nan
        if self.ins_mode in NOTCH_FILTER_RANGES:
            lmin, lmax = self.notch_limits
            cube[lmin:lmax + 1] = np.nan
        return cube

    def remold(self):
        """ Subtracts the reconstructed residuals and places the cleaned
        spectra into the duplicated datacube.
        """
        logger.info('Applying correction and reshaping data product')
        self.cleancube = self.make_cube_from_stack(self.stack - self.recon,
                                                   with_nans=self.run_clean)

    def reprocess(self, nevals=[]):
        """ A method that redoes the eigenvalue selection, reconstruction, and
        remolding of the data.
        """
        self.chooseevals(nevals=nevals)
        self.reconstruct()
        self.remold()

    def optimize(self):
        """Compute the optimal number of components needed to characterize
        the residuals.
        This function calculates the variance per segment with an increasing
        number of eigenspectra/eigenvalues. It then deterimines the point at
        which the second derivative of this variance curve reaches zero. When
        this occurs, the linear reduction in variance is attributable to the
        removal of astronomical features rather than emission line residuals.
        """
        logger.info('Compute number of components')
        ncomp = []
        for model in self.models:
            var = model.explained_variance_
            deriv, mn1, std1 = _compute_deriv(var)
            cross = np.append([False], deriv >= (mn1 - std1))
            ncomp.append(np.where(cross)[0][0])

        self.nevals = np.array(ncomp)

    def make_contcube(self):
        """ Remold the continuum array so it can be investigated.
        Takes the continuum stack and returns it into a familiar cube form.
        """
        contcube = self.cube.copy() * np.nan
        contcube[:, self.y, self.x] = self.contarray
        return contcube

    def _applymask(self, mask):
        """Apply a mask to the input data to provide a cleaner basis set.
        mask is >1 for objects, 0 for sky so that people can use sextractor.
        The file is read with ``astropy.io.fits.getdata`` which first tries to
        read the primary extension, then the first extension is no data was
        found before.
        """
        logger.info('Applying Mask for SVD Calculation from %s', mask)
        self.maskfile = mask
        mask = fits.getdata(mask).astype(bool)
        nmasked = np.count_nonzero(mask)
        logger.info('Masking %d pixels (%d%%)', nmasked,
                    nmasked / np.prod(mask.shape) * 100)
        self.cube[:, mask] = np.nan

    def writecube(self, outcubefits='DATACUBE_ZAP.fits', overwrite=False):
        """Write the processed datacube to an individual fits file."""
        outhead = _newheader(self)
        outhdu = fits.PrimaryHDU(data=self.cleancube, header=outhead)
        outhdu.writeto(outcubefits, overwrite=overwrite)
        logger.info('Cube file saved to %s', outcubefits)

    def writeskycube(self, skycubefits='SKYCUBE_ZAP.fits', overwrite=False):
        """Write the processed datacube to an individual fits file."""
        outcube = self.cube - self.cleancube
        outhead = _newheader(self)
        outhdu = fits.PrimaryHDU(data=outcube, header=outhead)
        outhdu.writeto(skycubefits, overwrite=overwrite)
        logger.info('Sky cube file saved to %s', skycubefits)

    def writevarcurve(self, varcurvefits='VARCURVE_ZAP.fits', overwrite=False):
        """Write the explained variance curves to an individual fits file."""
        from astropy.table import Table
        table = Table([m.explained_variance_ for m in self.models])
        hdu = fits.table_to_hdu(table)
        _newheader(self, hdu.header)
        hdu.writeto(varcurvefits, overwrite=overwrite)
        logger.info('Variance curve file saved to %s', varcurvefits)

    def mergefits(self, outcubefits, overwrite=False):
        """Merge the ZAP cube into the full muse datacube and write."""
        # make sure it has the right extension
        outcubefits = outcubefits.split('.fits')[0] + '.fits'
        with fits.open(self.cubefits) as hdu:
            if self.instrument == 'MUSE':
                hdu[1].header = _newheader(self)
                hdu[1].data = self.cleancube
            elif self.instrument == 'KCWI':
                hdu[0].header = _newheader(self)
                hdu[0].data = self.cleancube
            else:
                raise ValueError('unsupported instrument %s' % self.instrument)

            hdu.writeto(outcubefits, overwrite=overwrite)
        logger.info('Cube file saved to %s', outcubefits)

    def plotvarcurve(self, i=0, ax=None):
        var = self.models[i].explained_variance_
        deriv, mn1, std1 = _compute_deriv(var)

        if ax is None:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(3, 1, figsize=[10, 15])

        ax1, ax2, ax3 = ax
        ax1.plot(var, linewidth=3)
        ax1.plot([self.nevals[i], self.nevals[i]], [min(var), max(var)])
        ax1.set_ylabel('Variance')

        ax2.plot(np.arange(deriv.size), deriv)
        ax2.hlines([mn1, mn1 - std1], 0, len(deriv), colors=('k', '0.5'))
        ax2.plot([self.nevals[i] - 1, self.nevals[i] - 1],
                 [min(deriv), max(deriv)])
        ax2.set_ylabel('d/dn Var')

        deriv2 = np.diff(deriv)
        ax3.plot(np.arange(deriv2.size), np.abs(deriv2))
        ax3.plot([self.nevals[i] - 2, self.nevals[i] - 2],
                 [min(deriv2), max(deriv2)])
        ax3.set_ylabel('(d^2/dn^2) Var')
        # ax3.set_xlabel('Number of Components')

        ax1.set_title('Segment {0}, {1} - {2} Angstroms'.format(
            i, self.lranges[i][0], self.lranges[i][1]))

    def plotvarcurves(self):
        nseg = len(self.models)
        import matplotlib.pyplot as plt
        fig, axes = plt.subplots(nseg, 3, figsize=(16, nseg * 2),
                                 tight_layout=True)
        for i in range(nseg):
            self.plotvarcurve(i=i, ax=axes[i])


 # ================= Helper Functions =================

 def worker(f, i, chunk, out_q, err_q, kwargs):
    try:
        result = f(i, chunk, **kwargs)
    except Exception as e:
        err_q.put(e)
        return

    # output the result and task ID to output queue
    out_q.put((i, result))


 def parallel_map(func, arr, indices, **kwargs):
    logger.debug('Running function %s with %s chunks', func.__name__, indices)
    axis = kwargs.pop('axis', None)
    if isinstance(indices, (int, np.integer)) and indices == 1:
        return [func(0, arr, **kwargs)]

    manager = Manager()
    out_q = manager.Queue()
    err_q = manager.Queue()
    jobs = []
    chunks = np.array_split(arr, indices, axis=axis)
    if 'split_arrays' in kwargs:
        split_arrays = [np.array_split(a, indices, axis=axis)
                        for a in kwargs.pop('split_arrays')]
    else:
        split_arrays = None

    for i, chunk in enumerate(chunks):
        if split_arrays:
            kwargs['split_arrays'] = [s[i] for s in split_arrays]
        p = Process(target=worker, args=(func, i, chunk, out_q, err_q, kwargs))
        jobs.append(p)
        p.start()

    # gather the results
    for proc in jobs:
        proc.join()

    if not err_q.empty():
        # kill all on any exception from any one slave
        raise err_q.get()

    # Processes finish in arbitrary order. Process IDs double
    # as index in the resultant array.
    results = [None] * len(jobs)
    while not out_q.empty():
        idx, result = out_q.get()
        results[idx] = result

    return results


 def _compute_deriv(arr, nsigma=5):
    """Compute statistics on the derivatives"""
    npix = int(0.25 * arr.shape[0])
    deriv = np.diff(arr[:npix])
    ind = int(.15 * deriv.size)
    mn1 = deriv[ind:].mean()
    std1 = deriv[ind:].std() * nsigma
    return deriv, mn1, std1


 def _continuumfilter(stack, cftype, cfwidth=300, notch_limits=None):
    if cftype == 'fit':
        x = np.arange(stack.shape[0])

        # Excluding the very red part for the fit. This is Muse-specific,
        # but anyway for another instrument this method should probably
        # not be used as is.
        w = np.ones(stack.shape[0])
        w[3600:] = 0

        if notch_limits is not None:
            # Exclude the notch filter region
            lmin, lmax = notch_limits
            w[lmin:lmax + 1] = 0

        res = np.polynomial.polynomial.polyfit(x, stack, deg=5, w=w)
        ret = np.polynomial.polynomial.polyval(x, res, tensor=True)
        return ret.T

    if cftype == 'median':
        func = _icfmedian
    else:
        raise ValueError('unknown cftype option')

    logger.info('Using cfwidth=%d', cfwidth)

    if notch_limits is not None:
        # To manage the notch filter which is filled with zeros, we process the
        # stack in two halves, before and after the filter.
        c = np.zeros_like(stack)
        ctmp = parallel_map(func, stack[:notch_limits[0]], NCPU, axis=1,
                            cfwidth=cfwidth)
        c[:notch_limits[0]] = np.concatenate(ctmp, axis=1)
        ctmp = parallel_map(func, stack[notch_limits[1]:], NCPU, axis=1,
                            cfwidth=cfwidth)
        c[notch_limits[1]:] = np.concatenate(ctmp, axis=1)
    else:
        c = parallel_map(func, stack, NCPU, axis=1, cfwidth=cfwidth)
        c = np.concatenate(c, axis=1)

    return c


 def _icfmedian(i, stack, cfwidth=None):
    ufilt = 3  # set this to help with extreme over/under corrections
    return ndi.median_filter(
        ndi.uniform_filter(stack, (ufilt, 1)), (cfwidth, 1))


 def rolling_window(a, window):  # function for striding to help speed up
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


 def _newheader(zobj, header=None):
    """Put the pertinent zap parameters into the header"""
    header = header or zobj.header.copy()
    header['COMMENT'] = 'These data have been ZAPped!'
    header.append(('ZAPvers', __version__, 'ZAP version'), end=True)
    # zlevel removal performed
    header.append(('ZAPzlvl', zobj.run_zlevel, 'ZAP zero level correction'))
    # Nanclean performed
    header['ZAPclean'] = (zobj.run_clean,
                          'ZAP NaN cleaning performed for calculation')
    # Continuum Filtering
    header['ZAPcftyp'] = (zobj._cftype, 'ZAP continuum filter type')
    header['ZAPcfwid'] = (zobj._cfwidth, 'ZAP continuum filter size')

    # number of segments
    nseg = len(zobj.pranges)
    header['ZAPnseg'] = (nseg, 'Number of segments used for ZAP SVD')

    # per segment variables
    if hasattr(zobj, 'nevals'):
        for i in range(nseg):
            header['ZAPseg{}'.format(i)] = (
                '{}:{}'.format(zobj.pranges[i][0], zobj.pranges[i][1] - 1),
                'spectrum segment (pixels)')
            header['ZAPnev{}'.format(i)] = (zobj.nevals[i],
                                            'number of eigenvals/spectra used')

    return header


 def _isigclip(i, istack):
    mn = []
    for col in istack:
        clipped, bot, top = sigmaclip(col, low=3, high=3)
        mn.append(clipped.mean())
    return np.array(mn)


 def _imedian(i, istack):
    return np.median(istack, axis=1)


 @timeit
 def _nanclean(cube, rejectratio=0.25, boxsz=1):
    """
    Detects NaN values in cube and removes them by replacing them with an
    interpolation of the nearest neighbors in the data cube. The positions in
    the cube are retained in nancube for later remasking.
    """
    logger.info('Cleaning NaN values in the cube')
    cleancube = cube.copy()
    badcube = np.logical_not(np.isfinite(cleancube))        # find NaNs
    badmap = badcube.sum(axis=0)  # map of total nans in a spaxel

    # choose some maximum number of bad pixels in the spaxel and extract
    # positions
    badmask = badmap > (rejectratio * cleancube.shape[0])
    logger.info('Rejected %d spaxels with more than %.1f%% NaN pixels',
                np.count_nonzero(badmask), rejectratio * 100)

    # make cube mask of bad spaxels
    badcube &= (~badmask[np.newaxis, :, :])
    z, y, x = np.where(badcube)

    neighbor = np.zeros((z.size, (2 * boxsz + 1)**3))
    icounter = 0
    logger.info("Fixing %d remaining NaN pixels", len(z))

    # loop over samplecubes
    nz, ny, nx = cleancube.shape
    for j in range(-boxsz, boxsz + 1, 1):
        for k in range(-boxsz, boxsz + 1, 1):
            for l in range(-boxsz, boxsz + 1, 1):
                iz, iy, ix = z + l, y + k, x + j
                outsider = ((ix <= 0) | (ix >= nx - 1) |
                            (iy <= 0) | (iy >= ny - 1) |
                            (iz <= 0) | (iz >= nz - 1))
                ins = ~outsider
                neighbor[ins, icounter] = cleancube[iz[ins], iy[ins], ix[ins]]
                neighbor[outsider, icounter] = np.nan
                icounter = icounter + 1

    cleancube[z, y, x] = np.nanmean(neighbor, axis=1)
    return cleancube, badcube

except:
 failure=1


class MuseZap(esorexplugin.RecipePlugin):
    # The recipe must be given a name by setting the 'name' class attribute.
    name = "muse_zap"

    # A version needs to be assigned. For this it is best to use the
    # VersionNumber method that will correctly encode a major, minor and micro
    # version sub-numbers. NOTE: minor and micro must be in the range [0..99].
    version = VersionNumber(2, 0, 0)

    # A synopsis string should be given for the recipe for the man page created
    # by EsoRex.
    # NOTE: If the following class attribute is not provided, the first line of
    # this class's doc string will be used for the synopsis instead.
    synopsis = "This recipe performs the ZAP v2.1 algorithm to remove the residual sky lines in MUSE datacubes."

    description = ("This recipe performs the ZAP v2.1 algorithm to remove the residual sky lines in MUSE datacubes\n"
                   "It is part of a series of contributed recipes and it is meant to be used within the muse_zap \n"
                   "ESOreflex workflow. It is not part of the official MUSE pipeline.\n"
                   "For a proper description of the ZAP algorithm, we refer to the article Soto et al. 2016, MNRAS, 458, 3210S.\n"
                   "For any question related to this recipe, please contact sdp_muse@eso.org.\n")
    author = "Lodovico Coccato"
    email = "sdp_muse@eso.org"
    copyright = "Copyright 2017"


    names=["zlevel",
           "cftype",
           "cfwidthSVD",
           "cfwidthSP",
           "nevals",
           "clean"]#,
    #       "test"]
    #           "extSVD"]

    alias=["zlevel",
           "cftype",
           "cfwidthSVD",
           "cfwidthSP",
           "nevals",
           "clean"]#,
    #       "test"]
    #           "extSVD"]

    fullnames = list(map(lambda x: "muse_zap.muse_zap."+x, names))

    desc1="Method for the zeroth order sky removal: 'none', 'sigclip' or 'median' (default)."
    desc2="Method for the continuum filter: 'median' or 'weight' (default). For the 'weight' method, a zeroth order sky is required (see 'zlevel')."
    desc3="Window size for the continuum filter, for the SVD computation. Default to 500."
    desc4= "Window size for the continuum filter used to remove the continuum features for calculating the eigenvalues per spectrum. Smaller values better trace the sources. An optimal range of is typically 20 - 50 pixels. Default to 30."


        #Provide either a single value
        #that will be used for all of the segments, or a list of 11 values that
        #will be used for each of the segments."

    desc7= "Allow to specify the number of eigenspectra used for each segment. Ignored if pevals is used or optimizetType different from None."
        #Provide either a single value
        #that will be used for all of the segments, or a list of 11 values that
        #will be used for each of the segments.

    desc8= "If True (default value), the NaN values are cleaned. Spaxels with more then 25% of NaN values are removed, the others are replaced with an interpolation from the neighbors. The NaN values are reinserted into the final datacube. If set to False, any spaxel with a NaN value will be ignored."

#    desc9="Path of an input FITS file containing a single value decomposition (SVD) previously computed and to be used on the current object. If not provided, the SVD are computed from the imput frame"


 #   desc_test="test parameter"
    parameters = [
        EnumParameter(fullnames[0],'median', ['median', 'none', 'sigclip'],cli_alias=alias[0],cli_enabled=True,description=desc1),
        EnumParameter(fullnames[1],'median', ['fit', 'median','none'],cli_alias=alias[1],cli_enabled=True,description=desc2),
        ValueParameter(fullnames[2],500,cli_alias=alias[2],cli_enabled=True, description=desc3),
        ValueParameter(fullnames[3],30,cli_alias=alias[3],cli_enabled=True, description=desc4),
        ValueParameter(fullnames[4],-1  ,cli_alias=alias[4],cli_enabled=True, description=desc7),
        ValueParameter(fullnames[5],True,cli_alias=alias[5],cli_enabled=True, description=desc8)]
      ##  EnumParameter(fullnames[6],1.0,[0.,1.0,10.],cli_alias=alias[6],cli_enabled=True, description=desc_test)]

    def set_frame_group(self, frame):
        if frame.tag == 'RAW':
            frame.group = Frame.GROUP_RAW
        else:
            frame.group = Frame.GROUP_CALIB
        frame.level = Frame.LEVEL_FINAL

    def process(self,frames,zlevel,cftype,cfwidthSVD,cfwidthSP,nevals,clean):#zlevel,cftype,cfwidthSVD,cfwidthSP,optimizeType,pevals,nevals,clean,extSVD):

#if __name__ == '__main__':
        warnings.filterwarnings("ignore")
        output_frames = []
        values=[zlevel,cftype,cfwidthSVD,cfwidthSP,nevals,clean]

        mask_cube_filename=None
        mask_sky_filename=None
        IsThereSkyCube = 0
        IsThereObjectCube = 0

        for j in range(0,len(frames)):
            if frames[j].tag == 'MASK_FINAL_CUBE':
                mask_filename = frames[j].filename
            if frames[j].tag == 'DATACUBE_SKY':
                IsThereSkyCube = 1
                muse_sky_cubefits_filename = frames[j].filename
            if frames[j].tag == 'DATACUBE_OBJECT':
                IsThereObjectCube = 1
                muse_object_cubefits_filename = frames[j].filename


        outcubefits_filename = 'DATACUBE_CLEANED.fits'
        if IsThereSkyCube ==1:
          IsThereObjectCube = 0
          extSVD = SVDoutput(muse_sky_cubefits_filename, mask=mask_filename,
                  clean=clean, zlevel=zlevel, cftype=cftype,
                  cfwidth=cfwidthSVD,ncpu=NCPU)


          try:
            if nevals == -1:
               nevals = []

            processzap(muse_object_cubefits_filename, outcubefits=outcubefits_filename,
                       clean=clean,zlevel=zlevel,cftype=cftype, cfwidthSVD=cfwidthSVD,
                       cfwidthSP=cfwidthSP,  nevals=nevals,ncpu=NCPU,
                       extSVD=extSVD, skycubefits=None, interactive=False)

          except:
            self.raise_error(' execution error, check if products already exist')


        if IsThereObjectCube ==1:

          try:
             if nevals == -1:
                nevals = []
             processzap(muse_object_cubefits_filename, outcubefits=outcubefits_filename,clean=clean,zlevel=zlevel,
                              cftype=cftype, cfwidthSVD=cfwidthSVD, cfwidthSP=cfwidthSP, nevals=nevals,ncpu=NCPU,
                              mask=mask_filename, interactive=False)

          except:
                self.raise_error(' execution error, check if products already exist')



        hdu = fits.open(outcubefits_filename,method='update')
        hdu[0].header['HIERARCH ESO PRO CATG'] = 'DATACUBE_CLEANED'
        if IsThereSkyCube ==1:
            s1=muse_sky_cubefits_filename.split("/")[-1]
            s2=mask_filename.split("/")[-1]
            s1=s1[:50]
            s2=s2[:50]
            hdu[0].header['HIERARCH ESO SKYCUBE'] = s1
            hdu[0].header['HIERARCH ESO SKYCUBE MASK'] = s2
        else:
            hdu[0].header['HIERARCH ESO SKYCUBE'] = 'None'
            hdu[0].header['HIERARCH ESO SKYCUBE MASK'] = 'None'
        if IsThereObjectCube ==1:
            s3 = mask_filename.split("/")[-1]
            s3=s3[:50]
            hdu[0].header['HIERARCH ESO OBJECTCUBE MASK'] = s3
        else:
            hdu[0].header['HIERARCH ESO OBJECTCUBE MASK'] = 'None'

        for k in range(len(self.input_parameters)):
            hkey1 = 'HIERARCH ESO ZAP REC'+str(k+1)+' NAME'
            hkey2 = 'HIERARCH ESO ZAPn REC'+str(k+1)+' VALUE'
            hdu[0].header[hkey1] = self.alias[k]
            hdu[0].header[hkey2] = values[k]

        outcubefits = Frame(outcubefits_filename,'DATACUBE_CLEANED', type = Frame.TYPE_IMAGE)
        output_frames.append(outcubefits)

        outcubefits.write(hdu,overwrite = True, output_verify = 'ignore',checksum=True)



        return output_frames






try:
 import esorexplugin
 from esorexplugin import *
 from astropy.io import fits
except ImportError:
    import_ok = 0

import logging    
import sys   
import numpy as np 

try:
  from astropy.io import fits as pyfits
  from scipy.ndimage.filters import minimum_filter
except ImportError:
    import_ok = 0

 
logging.basicConfig(format='[%(levelname)s] %(message)s', level=logging.INFO,
                    stream=sys.stdout)
logger = logging.getLogger(__name__)
 
# Must define a new class deriving from esorexplugin.RecipePlugin to write a
# new Python based recipe plugin for EsoRex.

def crea_mask(image,min,max):

        try:
            where_are_NaNs = np.isnan(image)
            
            vector = image[~where_are_NaNs]
            num = len(vector)
            vector.sort()
            sel = vector[int(num*min) : int(num*max)]
            min_sel = sel.min()
            max_sel = sel.max()
            mask =  np.where((image >=min_sel) & (image<= max_sel) ,0,1)
            
        except:
            logger.warning('could not apply limits')
            mask= image * 0 + 1
            
        mask[where_are_NaNs] = 1
        
        return mask


def add_neibors(image,n_grow):
        mask= minimum_filter(image, size=n_grow)
        return mask
    
def add_skyboxes(image,skyboxes):
        ydim= len(image)
        xdim= len(image[0])
        for skybox in skyboxes:
            try:
                junk = skybox.replace('"', ' ')
                xmin, xmax, ymin, ymax = map(int,junk.split())
                if xmin >= 0 and xmax <= xdim and ymin >= 0 and ymax <= ydim:
                    image[ymin:ymax,xmin:xmax]= 0
            except:
                self.raise_error("can't use skybox " + skybox)
  
try:
 class CreaMaskRecipe(esorexplugin.RecipePlugin):

    # The recipe must be given a name by setting the 'name' class attribute.
    name = "muse_zap_crea_mask"     

    # A version needs to be assigned. For this it is best to use the
    # VersionNumber method that will correctly encode a major, minor and micro
    # version sub-numbers. NOTE: minor and micro must be in the range [0..99].
    version = VersionNumber(2, 0, 0)

    # A synopsis string should be given for the recipe for the man page created
    # by EsoRex.
    # NOTE: If the following class attribute is not provided, the first line of
    # this class's doc string will be used for the synopsis instead.
    synopsis = "Create or modify sky mask; idendify the sky regions to compute principal components"

    # A detailed description string should be given for the recipe for the man
    # page created by EsoRex.
    # NOTE: If the following class attribute is not provided, the whole text in
    # the class's doc string will be used for the description instead.
    description = ("Create or modify sky mask; idendify the sky regions to compute principal components"
                   "This recipe It is part of a series of contributed recipes and it is meant to be used within the muse_zap \n"
                   "ESOreflex workflow. It is not part of the official MUSE pipeline.\n"
                   "For any question related to this recipe, please contact sdp_muse@eso.org.\n")
    # The following additional fields should be provided for the man page
    # generated by EsoRex.
    author = "Lodovico Coccato"
    email = "sdp_muse@eso.org"
    
    copyright = EsoCopyright("MUSE_ZAP", [2010, 2013, 2014, 2015, 2017])

 
    names=["min_frac",
           "max_frac",
           "n_grow",
           "use_input_mask",
           "sky_region_1",
           "sky_region_2",
           "sky_region_3",
           "sky_region_4",
           "sky_region_5",
           "sky_region_6",
           "sky_region_7",
           "sky_region_8",
           "sky_region_9",
           "sky_region_10"]

    alias=["min_frac",
           "max_frac",
           "n_grow",
           "use_input_mask",
           "sky_region_1",
           "sky_region_2",
           "sky_region_3",
           "sky_region_4",
           "sky_region_5",
           "sky_region_6",
           "sky_region_7",
           "sky_region_8",
           "sky_region_9",
           "sky_region_10"]
    fullnames = list(map(lambda x: "muse_zap.muse_zap_crea_mask."+x, names))
     
    desc1= "Fraction of the image to be ignored"
    desc2= "Fraction of the image (without the ignored part) to be considered as sky"
    desc3= "number of adjacent pixels about each sky pixel to add to sky region"
    desc4= "Specification how to use the input mask if provided: 'use'= use as is; 'add'= use it and add to it according to parameters n_grow and skyregions to mask; 'discard'= use parameters n_grow and skyboxes to create completely new mask"
    desc5= "sky region 1 specified as xmin xmax ymin ymax"
    desc6= "sky region 2 specified as xmin xmax ymin ymax"
    desc7= "sky region 3 specified as xmin xmax ymin ymax"
    desc8= "sky region 4 specified as xmin xmax ymin ymax"
    desc9= "sky region 5 specified as xmin xmax ymin ymax"
    desc10= "sky region 6 specified as xmin xmax ymin ymax"
    desc11= "sky region 7 specified as xmin xmax ymin ymax"
    desc12= "sky region 8 specified as xmin xmax ymin ymax"
    desc13= "sky region 9 specified as xmin xmax ymin ymax"
    desc14= "sky region 10 specified as xmin xmax ymin ymax"

   
    #names=map(lambda x: 'muse_zap.muse_zap_crea_mask.'+x,alias)
    parameters = [
        RangeParameter(fullnames[0],0.1,0.0,1.0,cli_alias=alias[0],cli_enabled=True, description=desc1),
        RangeParameter(fullnames[1],0.7,0.0,1.0,cli_alias=alias[1],cli_enabled=True, description=desc2),
        ValueParameter(fullnames[2],3,cli_alias=alias[2],cli_enabled=True, description=desc3),
        EnumParameter(fullnames[3], "use", ["use", "add", "discard"],cli_alias=alias[3],cli_enabled=True, description=desc4),
        ValueParameter(fullnames[4],'-1 -1 -1 -1',cli_alias=alias[4],cli_enabled=True, description=desc5),
        ValueParameter(fullnames[5],'-1 -1 -1 -1',cli_alias=alias[5],cli_enabled=True, description=desc6),
        ValueParameter(fullnames[6],'-1 -1 -1 -1',cli_alias=alias[6],cli_enabled=True, description=desc7),
        ValueParameter(fullnames[7],'-1 -1 -1 -1',cli_alias=alias[7],cli_enabled=True, description=desc8),
        ValueParameter(fullnames[8],'-1 -1 -1 -1',cli_alias=alias[8],cli_enabled=True, description=desc9),
        ValueParameter(fullnames[9],'-1 -1 -1 -1',cli_alias=alias[9],cli_enabled=True, description=desc10),
        ValueParameter(fullnames[10],'-1 -1 -1 -1',cli_alias=alias[10],cli_enabled=True, description=desc11),
        ValueParameter(fullnames[11],'-1 -1 -1 -1',cli_alias=alias[11],cli_enabled=True, description=desc12),
        ValueParameter(fullnames[12],'-1 -1 -1 -1',cli_alias=alias[12],cli_enabled=True, description=desc13),
        ValueParameter(fullnames[13],'-1 -1 -1 -1',cli_alias=alias[13],cli_enabled=True, description=desc14)
        ]
 
    # The set_frame_group method must be overriden by a new recipe plugin. The
    # purpose of this method is to catalogue the input frames to the recipe.
    # Primarily to indicate to EsoRex if an input frame is a raw frame or
    # calibration frame.



    def set_frame_group(self, frame):
        if frame.tag == 'RAW':
            frame.group = Frame.GROUP_RAW
        else:
            frame.group = Frame.GROUP_CALIB
        frame.level = Frame.LEVEL_FINAL



    # A method called process must be defined that will take as first argument
    # 'frames', which is a list of Frame objects. Initially 'frames' contains
    # the input Frames from the SOF. As the process method executes it should
    # append any new output Frame objects to the 'frames' list.
    # The additional attributes should be the parameters as declared in the
    # 'parameters' class attribute above. There should be as many method
    # arguments as there were parameters declared and their names must
    # correspond.
#    def process(self, frames, par1, par2, par3):
  #  def process(self, frames,**kwargs):
    def process(self, frames,min_frac,max_frac,n_grow,use_input_mask,sky_region_1,
                sky_region_2,sky_region_3,sky_region_4,sky_region_5,
                sky_region_6,sky_region_7,sky_region_8,sky_region_9,sky_region_10):

        
        values=[min_frac,max_frac,n_grow,use_input_mask,sky_region_1,
                sky_region_2,sky_region_3,sky_region_4,sky_region_5,
                sky_region_6,sky_region_7,sky_region_8,sky_region_9,sky_region_10]
        
       

        # This will rarely be needed, but just in case, have more detailed
        # parameter information in self.input_parameters.
        # This is a list of RecipeParameter objects.
        # print(self.input_parameters)
        # self.input_parameters
        # Errors should be raised with the raise_error method.
        # Its signature is:
        #   def raise_error(self, message, exitcode = 1, print_traceback = False)
        if len(frames) < 1:
            self.raise_error("Need at least one input frame.")


        skyboxes= []

        if min_frac < 0:
            min_frac = 0.0
        if max_frac > 1:
            max_frac = 1.0
                    
        if min_frac < 0 and max_frac > 1:
            self.raise_error('Warning')
                        
        if min_frac < 0 and max_frac < 0:
            self.raise_error('Warning')
                            
        if min_frac > 1 and max_frac >1:
            self.raise_error('Warning')
                                
        if min_frac > max_frac:
            min_frac = 0.0
            max_frac = 0.0
            self.raise_error('Warning')
                                    
            
        skyboxes.append(sky_region_1)
        skyboxes.append(sky_region_2)
        skyboxes.append(sky_region_3)
        skyboxes.append(sky_region_4)
        skyboxes.append(sky_region_5)
        skyboxes.append(sky_region_6)
        skyboxes.append(sky_region_7)
        skyboxes.append(sky_region_8)
        skyboxes.append(sky_region_9)
        skyboxes.append(sky_region_10)

        bad_parameters = 0
        #read input images
        input_mask_provided = False
        for j in range(0,len(frames)):
            if frames[j].tag == 'IMAGE_OBJECT' or frames[j].tag == 'IMAGE_SKY':
                hdu_fov  = fits.open(frames[j].filename,checksum=True)
                image_fov = hdu_fov[1].data 
        #input_mask_provided = False
            if frames[j].tag == 'MASK_OBJECT' or frames[j].tag == 'MASK_SKY':
                input_mask_name = frames[j].filename#file.name
                hdu  = fits.open(input_mask_name,checksum=True)
                input_mask = hdu[0].data
                input_mask_provided = True

        if input_mask_provided and  use_input_mask[0] == 'u':
           input_mask_name = output_name = 'MASK_FINAL_CUBE.fits'
           output_mask = Frame(input_mask_name, "MASK_FINAL_CUBE")
           output_frames = []
           output_frames.append(output_mask)
           output_mask.write(hdu, overwrite = True, output_verify = 'ignore',checksum=True)
           
        else: 
           if input_mask_provided and  use_input_mask[0] == 'a':
                first_mask= input_mask
           else:
                first_mask = crea_mask(image_fov,min_frac,max_frac)
                mask = image_fov
           mask = add_neibors(first_mask, n_grow)
           add_skyboxes(mask, skyboxes)
 

           #check for bad parameters and create a fake mask + error message
           if np.min(mask) != 0 or np.max(mask) !=1:
              mask=mask*0.0+1.
              mask[0]= 0
              self.raise_error('Warning, invalid mask created. Check the recipe parameters')
           # create output mask:
           
           output_name = 'MASK_FINAL_CUBE_'+str(j+1)+'.fits'
           hdu2 = fits.PrimaryHDU(mask)
           hdulist = fits.HDUList([hdu2])
           hdulist[0].header = hdu_fov[0].header
           hdulist[0].header['HIERARCH ESO PRO CATG'] = 'MASK_FINAL_CUBE'
           hdulist[0].data=mask


           
           #Write recipe parameters in the output header
           for k in range(len(self.input_parameters)):
              
               hkey1 = 'HIERARCH ESO MASK REC'+str(k+1)+' NAME'
               hkey2 = 'HIERARCH ESO MASK REC'+str(k+1)+' VALUE'
               hdulist[0].header[hkey1] = self.alias[k]
               hdulist[0].header[hkey2] = values[k]
               
           output_mask = Frame(output_name, "MASK_FINAL_CUBE")

        # New output frames are created with the Frame object's constructor.
        # The signature for the constructor method is:
        #   def __init__(self, filename, tag, type = None, group = None, level = None)
       
        # Make sure to add the new frames to the 'frames' list to register them
        # with EsoRex.
           output_frames = []
           output_frames.append(output_mask)

        # One can write the associated FITS file with the frame's write method.
        # It accepts HDUList objects and additional.
           output_mask.write(hdulist, overwrite = True, output_verify = 'ignore',checksum=True)

        return output_frames
except:
 fail=1



from __future__ import with_statement
from __future__ import print_function

import sys
import os
import types

import_success = True
try:
    import numpy as np
    import reflex
    import pipeline_display
    import pipeline_product
    import reflex_plot_widgets as reflex_widgets
except ImportError:
    import_success = False
    print("Error importing reflex modules, numpy")


def get_inherited_key(fits_datacontainer, fits_dataset, key):
    value = None
    idx = None
    try:
        idx = fits_dataset
        value = fits_datacontainer.all_hdu[idx].header[key]
    except KeyError:
        # ESO DICB standard: Keyword inheritance is applied unless it is
        # explicitly disabled.
        inherit = True
        try:
            inherit = fits_datacontainer.all_hdu[idx].header["INHERIT"]
            if not isinstance(inherit, bool):
                inherit = True
        except KeyError:
            pass
        if inherit == True:
            try:
                idx = 0
                value = fits_datacontainer.all_hdu[idx].header[key]
            except KeyError:
                raise

    return (value, idx)


def format_coordinates(data, xpos, ypos):
    formatted_string = ""
    (img_data, wcs) = data

    xpix = int(xpos)
    ypix = int(ypos)
    x_inside = xpos > 0 and xpos <= img_data.image.shape[1]
    y_inside = ypos > 0 and ypos <= img_data.image.shape[0]
    if x_inside and y_inside:
        alpha, delta = wcs.image_to_celestial(xpos, ypos)
        formatted_string = ("%s alpha=%.4f delta=%.4f" +
                            "     Image X=%.3f  Y=%.3f" +
                            "     Value=%.3f") % \
            (wcs.radecsys, alpha, delta, xpos + 1., ypos + 1.,
             img_data.image[ypix, xpix])

    return formatted_string


def read_table(frame, columns, dataset=1):
    table = dict()
    for name in columns:
        try:
            table[name] = frame.all_hdu[dataset].data.field(name)
        except KeyError:
            table = None
            print("Warning: column '" + name + "' not found in " +
                  frame.fits_file.name)
        if not table:
            break
    return table


def cycle_marker_properties(index, colors, symbols):
    ncolors = len(colors)
    nsymbols = len(symbols)

    color = colors[index % ncolors]
    symbol = symbols[(index // ncolors) % nsymbols]
    return (symbol, color)


class WorldCoordinateSystem(object):
    def __init__(self, fits_datacontainer, fits_dataset=0, tiny=1.e-6):
        self.tiny = tiny
        self.ctype = None
        self.crval = None
        self.crpix = None
        self.cdelt = None
        self.pc_matrix = None
        self.cd_matrix = None

        try:
            _equinox = get_inherited_key(
                fits_datacontainer, fits_dataset, "EQUINOX")
        except KeyError:
            _equinox = None

        try:
            _radecsys = get_inherited_key(
                fits_datacontainer, fits_dataset, "RADESYS")
        except KeyError:
            try:
                _radecsys = fits_datacontainer.all_hdu[fits_dataset].header["RADECSYS"]
            except KeyError:
                _radecsys = None

        if _equinox is not None:
            equinox = _equinox[0]
        else:
            equinox = None

        if _radecsys is not None:
            radecsys = _radecsys[0]
        else:
            radecsys = None

        if equinox is not None:
            if radecsys is None:
                if equinox < 1984.0:
                    self.radecsys = "FK4"
                else:
                    self.radecsys = "FK5"
            else:
                self.radecsys = radecsys.strip()
        else:
            if radecsys is not None:
                self.radecsys = radecsys.strip()
                if self.radecsys == "FK4" or self.radecsys == "FK4-NO-E":
                    self.equinox = 1950.
                elif self.radecsys == "FK5":
                    self.equinox = 2000.
            else:
                self.radecsys = "ICRS"

        try:
            ctype1 = fits_datacontainer.all_hdu[fits_dataset].header["CTYPE1"]
            ctype2 = fits_datacontainer.all_hdu[fits_dataset].header["CTYPE2"]
            crval1 = fits_datacontainer.all_hdu[fits_dataset].header["CRVAL1"]
            crval2 = fits_datacontainer.all_hdu[fits_dataset].header["CRVAL2"]
            crpix1 = fits_datacontainer.all_hdu[fits_dataset].header["CRPIX1"]
            crpix2 = fits_datacontainer.all_hdu[fits_dataset].header["CRPIX2"]

            self.ctype = np.array([ctype1, ctype2])
            self.crval = np.array([crval1, crval2])
            self.crpix = np.array([crpix1, crpix2])

            if "CD1_1" in fits_datacontainer.all_hdu[fits_dataset].header:
                cd11 = fits_datacontainer.all_hdu[fits_dataset].header["CD1_1"]
                cd12 = fits_datacontainer.all_hdu[fits_dataset].header["CD1_2"]
                cd21 = fits_datacontainer.all_hdu[fits_dataset].header["CD2_1"]
                cd22 = fits_datacontainer.all_hdu[fits_dataset].header["CD2_2"]

                rho_a = 0.
                if cd21 > 0.:
                    rho_a = np.arctan2(cd21, cd11)
                elif cd21 < 0.:
                    rho_a = np.arctan2(-cd21, -cd11)
                rho_b = 0.
                if cd12 > 0.:
                    rho_b = np.arctan2(cd12, -cd22)
                elif cd12 < 0.:
                    rho_b = np.arctan2(-cd12, cd22)
                if np.fabs(rho_a - rho_b) < self.tiny:
                    rho = 0.5 * (rho_a + rho_b)
                    pc11 = np.cos(rho)
                    pc12 = -np.sin(rho)
                    if cd11 != 0.:
                        cdelt1 = cd11 / pc11
                    else:
                        cdelt1 = -cd21 / pc12
                    if cd22 != 0.:
                        cdelt2 = cd22 / pc11
                    else:
                        cdelt2 = cd12 / pc12
                else:
                    raise ValueError

                self.cdelt = np.array([[cdelt1, 0.], [0., cdelt2]])
                self.pc_matrix = np.array([[pc11, pc12], [-pc12, pc11]])
                self.cd_matrix = np.array([[cd11, cd12], [cd21, cd22]])
            else:
                cdelt1 = fits_datacontainer.all_hdu[fits_dataset].header["CDELT1"]
                cdelt2 = fits_datacontainer.all_hdu[fits_dataset].header["CDELT2"]

                pc11 = fits_datacontainer.all_hdu[fits_dataset].header["PC1_1"]
                pc12 = fits_datacontainer.all_hdu[fits_dataset].header["PC1_2"]
                pc21 = fits_datacontainer.all_hdu[fits_dataset].header["PC2_1"]
                pc22 = fits_datacontainer.all_hdu[fits_dataset].header["PC2_2"]

                self.cdelt = np.array([[cdelt1, 0.], [0., cdelt2]])
                self.pc_matrix = np.array([[pc11, pc12], [pc21, pc22]])
                self.cd_matrix = np.dot(self.cdelt, self.pc_matrix)

        except(KeyError, ValueError):
            self.ctype = None
            self.crval = None
            self.crpix = None
            self.cdelt = None
            self.cd_matrix = None
            self.pc_matrix = None

    def image_to_celestial(self, xpix, ypix):
        # Linear transformation
        pos = np.array([xpix - self.crpix[0], ypix -
                        self.crpix[1]]).transpose()
        xpos, ypos = np.dot(self.cd_matrix, pos)

        # Gnomonic projection
        phi = np.arctan2(xpos, -ypos)
        theta = np.arctan(1. / np.radians(np.sqrt(xpos**2. + ypos**2.)))

        alpha_p = np.radians(self.crval[0])
        delta_p = np.radians(self.crval[1])

        stheta = np.sin(theta)
        ctheta = np.cos(theta)
        sphi = np.sin(phi)
        cphi = np.cos(phi)
        sdelta_p = np.sin(delta_p)
        cdelta_p = np.cos(delta_p)

        ra_rad = alpha_p + np.arctan2(ctheta * sphi,
                                      stheta * cdelta_p +
                                      ctheta * sdelta_p * cphi)
        dec_rad = np.arcsin(stheta * sdelta_p - ctheta * cdelta_p * cphi)

        return (np.degrees(ra_rad), np.degrees(dec_rad))

    def celestial_to_image(self, ra_deg, dec_deg):
        ra_rad = np.radians(ra_deg)
        dec_rad = np.radians(dec_deg)

        alpha_p = np.radians(self.crval[0])
        delta_p = np.radians(self.crval[1])

        sra = np.sin(ra_rad - alpha_p)
        cra = np.cos(ra_rad - alpha_p)
        sdec = np.sin(dec_rad)
        cdec = np.cos(dec_rad)
        sdelta_p = np.sin(delta_p)
        cdelta_p = np.cos(delta_p)

        phi = np.arctan2(-cdec * sra,
                         sdec * cdelta_p - cdec * sdelta_p * cra) + np.pi
        theta = np.arcsin(sdec * sdelta_p + cdec * cdelta_p * cra)
        r_theta = np.rad2deg(1.) / np.tan(theta)

        xpos = r_theta * np.sin(phi)
        ypos = -r_theta * np.cos(phi)

        invdet = 1. / np.linalg.det(self.cd_matrix)

        xpix = (self.cd_matrix[1, 1] * xpos -
                self.cd_matrix[0, 1] * ypos) * invdet + self.crpix[0]
        ypix = (self.cd_matrix[0, 0] * ypos -
                self.cd_matrix[1, 0] * xpos) * invdet + self.crpix[1]

        return (xpix, ypix)


class DataPlotterManager(object):

    recipeName = "muse_exp_align"
    tagInputData = "IMAGE_FOV"
    tagFovPreview = "PREVIEW_FOV"
    tagAuxiliaryData = "SOURCE_LIST"

    _src_list_column_ids = ["Id", "X", "Y", "RA",
                            "DEC", "RA_CORR", "DEC_CORR", "Flux"]

    _btn_label_all = "All detections (uncorrected)"
    _btn_label_all_corr = "All detections (corrected)"

    _obs_id = "DATE-OBS"

    _marker_symbols = ["o", "s", "D", "p", "*", "+", "x"]
    _marker_colors = ["#0000FF", "#00FF00", "#FF0000", "#00FFFF",
                      "#FF00FF", "#FFFF00", "#FFA500", "#800080"]

    def __init__(self):
        self._plt = dict()
        self._widgets = dict()
        self._image_selector_active = 0
        self._get_parameter_value = None
        self._plt_plotter = None
        self._plt_creator = None
        self._figure = None
        self._resize_cid = None

        self.parameters = list()
        self.frameset = dict()

        super(DataPlotterManager, self).__init__()
        return

    def setWindowTitle(self):
        return self.recipeName + " GUI"

    def setWindowHelp(self):
        return "Help for " + self.setWindowTitle()

    def setInteractiveParameters(self):
        self.parameters += [
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="rsearch",
                group="Offset Calculation",
                description="Search radius (in arcsec)"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="nbins",
                group="Offset Calculation",
                description="Number of bins of the 2D histogram"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="weight",
                group="Offset Calculation",
                description="Use weighting"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="fwhm",
                group="Source Detection",
                description="FWHM of the convolution filter"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="bpixdistance", group="Source Detection",
                description="Minimum distance from image border [pixels]"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="threshold",
                group="Source Detection",
                description="Initial threshold for detecting point sources"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="bkgignore",
                group="Source Detection",
                description="Fraction of the image to be ignored"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="bkgfraction",
                group="Source Detection",
                description="Fraction of the image (without the ignored " +
                "part) considered as background"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="step",
                group="Source Detection",
                description="Increment/decrement of the threshold value"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="iterations",
                group="Source Detection",
                description="Maximum number of iterations"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="srcmin",
                group="Source Detection",
                description="Minimum number of sources"),
            reflex.RecipeParameter(
                recipe=self.recipeName,
                displayName="srcmax",
                group="Source Detection",
                description="Maximum number of sources")
        ]
        return self.parameters

    def setCurrentParameterHelper(self, parameter_value_getter):
        self._get_parameter_value = parameter_value_getter
        return

    def readFitsData(self, fitsfiles):
        self.frameset = dict()
        for fitsfile in fitsfiles:
            if fitsfile.category not in self.frameset:
                self.frameset[fitsfile.category] = []

            datasets = pipeline_product.PipelineProduct(fitsfile)

            frame_id = os.path.basename(datasets.fits_file.name)
            if fitsfile.category == self.tagInputData:
                frame_id += "\n(" + DataPlotterManager._obs_id + ": "
                frame_id += datasets.readKeyword(
                    DataPlotterManager._obs_id) + ")"

            frame = dict()
            frame["label"] = frame_id
            frame["datasets"] = datasets
            self.frameset[fitsfile.category].append(frame)

        if self.tagInputData in self.frameset and self.tagAuxiliaryData in self.frameset:
            self._plt_creator = self._data_plot_create
            self._plt_plotter = self._data_plot_draw
        else:
            self._plt_creator = self._dummy_plot_create
            self._plt_plotter = self._dummy_plot_draw
        return

    def addSubplots(self, figure):
        self._figure = figure
        self._figure.clear()
        self._plt_creator()
        return

    def plotWidgets(self):

        if "dummy" not in self._plt:
            if "buttons" not in self._widgets:
                button_labels = []
                for frame in self.frameset[self.tagInputData]:
                    button_labels.append(frame["label"])
                button_labels.append(DataPlotterManager._btn_label_all)
                button_labels.append(DataPlotterManager._btn_label_all_corr)

                self._widgets["buttons"] = \
                    reflex_widgets.InteractiveRadioButtons(
                        self._plt["buttons"],
                        self.on_image_selected,
                        button_labels,
                        self._image_selector_active,
                        title="Image Selector")
                self._resize_cid = self._figure.canvas.mpl_connect("resize_event",
                                                                   self.on_resize)

                for btn_label in self._widgets["buttons"].rbuttons.labels:
                    btn_label.set_clip_on(True)

        widgets = list()
        for key in self._widgets:
            widgets.append(self._widgets[key])

        return widgets

    def plotProductsGraphics(self):
        self._plt_plotter()
        return

    # Event handler

    def on_image_selected(self, button_label):
        selection_changed = False
        if button_label == DataPlotterManager._btn_label_all:
            if self._image_selector_active != -1:
                self._image_selector_active = -1
                selection_changed = True
        elif button_label == DataPlotterManager._btn_label_all_corr:
            if self._image_selector_active != -2:
                self._image_selector_active = -2
                selection_changed = True
        else:
            for idx, frame in enumerate(self.frameset[self.tagInputData]):
                if frame["label"] == button_label and idx != self._image_selector_active:
                    self._image_selector_active = idx
                    selection_changed = True

        # Update image view only if the selection changed
        if selection_changed:
            self._data_plot_draw()

        return

    def on_resize(self, event):
        if hasattr(self, '_widgets') and "buttons" in self._widgets:
            bbox = self._plt["buttons"].get_position().get_points()
            btns = self._widgets["buttons"].rbuttons
            figure = event.canvas.figure

            wbox = bbox[:, 0].ptp()
            hbox = bbox[:, 1].ptp()
            hfig = figure.get_figheight()
            wfig = figure.get_figwidth()
            vscale = (wbox * wfig) / (hbox * hfig)

            width = btns.labels[0].get_fontsize()
            width /= (figure.get_dpi() * wbox * wfig)

            for circle in btns.circles:
                circle.width = width
                circle.height = width * vscale
        return

    # Implementation of the plot creator and plotter delegates to be used
    # if the required data is actually available.

    def _data_plot_create(self):
        self._plt["buttons"] = self._figure.add_axes([0.05, 0.05, 0.35, 0.90])
        self._plt["imageview"] = self._figure.add_axes(
            [0.55, 0.05, 0.40, 0.90])

        # Turn off coordinate display in the status bar for the radio
        # buttons widget
        self._plt["buttons"].format_coord = lambda x, y: ""

        return

    def _data_plot_draw(self):
        if self._image_selector_active >= 0:
            img_frame = self.frameset[self.tagInputData][self._image_selector_active]
            img_data = img_frame["datasets"]
            img_data.readImage(1)

            wcs = WorldCoordinateSystem(img_data, 1)
            if wcs.ctype[0].strip() != "RA---TAN" or wcs.ctype[1].strip() != "DEC--TAN":
                raise ValueError("Unsupported world coordinate system found in '%s'" %
                                 (img_data.fits_file.name,))

            tooltip = img_frame["label"]

            # Find matching source list for the current FOV image img_data
            tbl_data = None
            num_sources = 0
            for frame in self.frameset[self.tagAuxiliaryData]:
                timestamp = frame["datasets"].readKeyword(
                    DataPlotterManager._obs_id)
                if timestamp == img_data.readKeyword(DataPlotterManager._obs_id):
                    print("Reading Source list '" + frame["label"] +
                          "' for FOV image '" + img_data.fits_file.name + "'")
                    tbl_data = read_table(frame["datasets"],
                                          self._src_list_column_ids)
                    if tbl_data is not None:
                        num_sources = len(tbl_data["Id"])

            tooltip += ": " + str(num_sources) + " sources detected"

            xlabel = "X [pixel]"
            ylabel = "Y [pixel]"

            self._plt["imageview"].clear()

            imageview = pipeline_display.ImageDisplay()
            imageview.setAspect("equal")
            imageview.setCmap("gray")
            imageview.setZAutoLimits(img_data.image, None)
            imageview.setLabels(xlabel, ylabel)
            imageview.setXLinearWCSAxis(0., 1., 1.)
            imageview.setYLinearWCSAxis(0., 1., 1.)

            imageview.display(self._plt["imageview"],
                              "Exposure FOV", tooltip, img_data.image)

            # This must be called after imageview display, to override the internal
            # formatting by imageview.display. This is a hack!

            self._plt["imageview"].format_coord = types.MethodType(format_coordinates,
                                                                   (img_data, wcs))

            self._plt["imageview"].autoscale(enable=False, tight=True)

            # If the source table is not empty, plot the detected source
            # positions and their labels on top of the displayed image
            if num_sources > 0:
                xpos = tbl_data["X"]
                ypos = tbl_data["Y"]
                marker_size = np.zeros(num_sources)
                marker_size[:] = 8

                marker, color = cycle_marker_properties(self._image_selector_active,
                                                        DataPlotterManager._marker_colors,
                                                        DataPlotterManager._marker_symbols)

                self._plt["imageview"].scatter(xpos, ypos,
                                               facecolors="none", edgecolors=color,
                                               marker=marker, s=marker_size**2)

                for idx, src_id in enumerate(tbl_data["Id"]):
                    txt_offset = 0.5 * np.sqrt(2.) * marker_size[idx]
                    xydata = (xpos[idx], ypos[idx])
                    xytext = (txt_offset, -txt_offset)
                    self._plt["imageview"].annotate(str(src_id),
                                                    xy=xydata, xycoords="data",
                                                    xytext=xytext, textcoords="offset points",
                                                    ha="left", va="top", fontsize="x-small",
                                                    color=color, clip_on=True)
        else:
            # If it exists, use the FOV preview image to setup the image
            # display and get the WCS. If the preview image is not available
            # use the first input image as fallback.
            have_fov_preview = False
            if self.tagFovPreview in self.frameset:
                img_frame = self.frameset[self.tagFovPreview][0]
                have_fov_preview = True
            else:
                img_frame = self.frameset[self.tagInputData][0]
            img_data = img_frame["datasets"]
            img_data.readImage(1)

            # Show the FOV preview as background image only if the selected
            # overlay are the corrected source positions.
            show_bkg_image = have_fov_preview and \
                self._image_selector_active in [-2]

            wcs = WorldCoordinateSystem(img_data, 1)
            if wcs.ctype[0].strip() != "RA---TAN" or wcs.ctype[1].strip() != "DEC--TAN":
                raise ValueError("Unsupported world coordinate system found in '%s'" %
                                 (img_data.fits_file.name,))

            if show_bkg_image == True:
                bkg_image = img_data.image
            else:
                # Set the pixel data of the image to 0 to clear it so that no
                # background image is displayed.
                bkg_image = np.array(img_data.image, copy=True)
                bkg_image[:] = 0

            src_lists = list()
            for img_frame in self.frameset[self.tagInputData]:
                timestamp = img_frame["datasets"].readKeyword(
                    DataPlotterManager._obs_id)
                img_name = img_frame["datasets"].fits_file.name
                for tbl_frame in self.frameset[self.tagAuxiliaryData]:
                    if timestamp == tbl_frame["datasets"].readKeyword(DataPlotterManager._obs_id):
                        print("Reading Source list '" + tbl_frame["label"] +
                              "' for FOV image '" + img_name + "'")
                        tbl_data = read_table(tbl_frame["datasets"],
                                              self._src_list_column_ids)
                        src_lists.append(tbl_data)

            xlabel = "X [pixel]"
            ylabel = "Y [pixel]"
            col_name_x = None
            col_name_y = None
            if self._image_selector_active == -1:
                col_name_x = "RA"
                col_name_y = "DEC"
                tooltip = "Uncorrected positions"
            else:
                col_name_x = "RA_CORR"
                col_name_y = "DEC_CORR"
                tooltip = "Corrected positions"

            self._plt["imageview"].clear()

            imageview = pipeline_display.ImageDisplay()
            imageview.setAspect("equal")
            imageview.setCmap("gray")
            if show_bkg_image == True:
                imageview.setZAutoLimits(bkg_image, None)
            else:
                imageview.setZLimits([0., 1.])
            imageview.setLabels(xlabel, ylabel)
            imageview.setXLinearWCSAxis(0., 1., 1.)
            imageview.setYLinearWCSAxis(0., 1., 1.)
            imageview.display(self._plt["imageview"],
                              "Source Positions", tooltip, bkg_image)

            # This must be called after imageview display, to override the internal
            # formatting by imageview.display. This is a hack!

            self._plt["imageview"].format_coord = types.MethodType(format_coordinates,
                                                                   (img_data, wcs))

            self._plt["imageview"].autoscale(enable=False)

            for idx, src_list in enumerate(src_lists):
                num_sources = len(src_list[col_name_x])
                if num_sources > 0:
                    img_coordinates = np.empty((num_sources, 2))
                    sky_coordinates = zip(
                        src_list[col_name_x], src_list[col_name_y])
                    for i, (ra_deg, dec_deg) in enumerate(sky_coordinates):
                        img_coordinates[i] = wcs.celestial_to_image(
                            ra_deg, dec_deg)
                    marker_size = np.zeros(num_sources)
                    marker_size[:] = 8
                    marker, color = cycle_marker_properties(idx,
                                                            DataPlotterManager._marker_colors,
                                                            DataPlotterManager._marker_symbols)
                    self._plt["imageview"].scatter(img_coordinates[:, 0], img_coordinates[:, 1],
                                                   facecolors="none", edgecolors=color,
                                                   marker=marker, s=marker_size**2)

        return

    # Implementation of the dummy plot creator and plotter delegates
    # follows here.

    def _dummy_plot_create(self):
        self._plt["dummy"] = self._figure.add_subplot(1, 1, 1)
        return

    def _dummy_plot_draw(self):
        label = "Data not found!\n\nInput files should contain these types:\n%s\n%s" \
            % (self.tagInputData, self.tagAuxiliaryData)

        self._plt["dummy"].set_axis_off()
        self._plt["dummy"].text(0.5, 0.5, label, color="#11557c", fontsize=18,
                                horizontalalignment="center", verticalalignment="center",
                                alpha=0.25)
        self._plt["dummy"].tooltip = "No data found"
        return


if __name__ == "__main__":

    from reflex_interactive_app import PipelineInteractiveApp

    interactive_app = PipelineInteractiveApp(enable_init_sop=True)
    interactive_app.parse_args()

    if not import_success:
        interactive_app.setEnableGUI(False)

    if interactive_app.isGUIEnabled():
        dataPlotManager = DataPlotterManager()

        interactive_app.setPlotManager(dataPlotManager)
        interactive_app.showGUI()
    else:
        interactive_app.set_continue_mode()

    # NOTE: Do not remove this line! This prints the output, which is parsed
    #       by the Reflex PythonActor to get the results!
    interactive_app.print_outputs()

    sys.exit()



from __future__ import with_statement
import sys
import time

try:
    import numpy
    import reflex
    from pipeline_product import PipelineProduct
    from reflex_plot_widgets import *
    import pipeline_display
    import_success = True

except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")
    

def paragraph(text, width=None):
    """ wrap text string into paragraph
       text:  text to format, removes leading space and newlines
       width: if not None, wraps text, not recommended for tooltips as
              they are wrapped by wxWidgets by default
    """
    import textwrap
    if width is None:
        return textwrap.dedent(text).replace('\n', ' ').strip()
    else:
        return textwrap.fill(textwrap.dedent(text), width=width)


class DataPlotterManager(object):
    """
    This class must be added to the PipelineInteractiveApp with setPlotManager
    It must have following member functions which will be called by the app:
     - setInteractiveParameters(self)
     - readFitsData(self, fitsFiles):
     - addSubplots(self, figure):
     - plotProductsGraphics(self)
    Following members are optional:
     - setWindowHelp(self)
     - setWindowTitle(self)
     - setCurrentParameterHelper(self, helper)
    """

    # static members
    recipe_name = "muse_zap_crea_mask"
    mask_cat = "MASK_FINAL_CUBE"
    fov_cat_sky = "IMAGE_SKY"
    fov_cat_obj = "IMAGE_OBJECT"

    def setInteractiveParameters(self):
        """
        This function specifies which are the parameters that should be presented
        in the window to be edited.  Note that the parameter has to also be in the
        in_sop port (otherwise it won't appear in the window). The descriptions are
        used to show a tooltip. They should match one to one with the parameter
        list.
        """

        # the parameters we send to the GUI depends on whether the in_sof_rec_orig includes a mask or not:
        mask_was_provided= False
        for file in self.orig_files:
            if file.category[0:4] == "MASK":
                    mask_was_provided= True

        if mask_was_provided:
          parlist= [
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="use_input_mask",
                                   group="Mask", description="How to use the user supplied mask: use= use the mask without modification;  add= add n_grow and skyregions to mask (parameters are specified in the 'Flux Ranges' and 'Sky Regions' tabs); discard=discard mask and use parameters in the 'Flux Ranges' and 'Sky Regions' tabs to create a new mask."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="min_frac",
                                   group="Flux Ranges", description="Fraction of faintest pixel NOT to use as sky. If N is the total number of pixels, then the N*min_frac faintest pixels will NOT be considered as sky."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="max_frac",
                                   group="Flux Ranges", description="Fraction of faint pixels to use as sky. If N is the total number of pixels, then the N*max_frac faintest pixels will be candidates for being a sky pixel. Note the parameter min_frac that can be used to exclude the very faintest pixels. "),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="n_grow",
                                   group="Flux Ranges", description="Add a region of ngrow pixels about any pixel found as sky using the flux range method."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_1",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_2",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_3",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_4",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_5",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_6",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_7",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_8",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_9",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_10",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax.")
            ]
        else:
          parlist= [
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="min_frac",
                                   group="Flux Ranges", description="Fraction of faintest pixel NOT to use as sky"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="max_frac",
                                   group="Flux Ranges", description="Fraction of faint pixel to use as sky."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="n_grow",
                                   group="Flux Ranges", description="Add a region of ngrow pixels about any pixel found as sky using the flux percentile method."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_1",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_2",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_3",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_4",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_5",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_6",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_7",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_8",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_9",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax."),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="sky_region_10",
                                   group="Sky Regions", description="Sky box specified as xmin xmax ymin ymax.")
            ]
        return parlist

    def readFitsData(self, fitsFiles):
        """
        This function should be used to read and organize the raw fits files
        produced by the recipes.
        It receives as input a list of reflex.FitsFiles
        """
        # organize the files into a dictionary, here we assume we only have 
        # one file per category if there are more, one must use a
        # dictionary of lists
        self.frames = dict()
        for f in fitsFiles:
            self.frames[f.category] = PipelineProduct(f)

        # we only have two states, we have data or we don't
        # define the plotting functions we want to use for each
        if (self.mask_cat in self.frames and self.fov_cat_sky in self.frames) or (self.mask_cat in self.frames and self.fov_cat_obj in self.frames):
            self._add_subplots = self._add_subplots
            self._plot = self._data_plot
        else:
            self._add_subplots = self._add_nodata_subplots
            self._plot = self._nodata_plot

    def addSubplots(self, figure):
        """
        This function should be used to setup the subplots of the gui.  The the
        matplotlib documentation for a description of subplots.
        """
        self._add_subplots(figure)

    def plotProductsGraphics(self):
        """
        This function should be used to plot the data onto the subplots.
        """
        self._plot()

    def setWindowHelp(self):
        return 'Help for mask interactive window'

    def setWindowTitle(self):
        return 'mask creation interactive window'

    def _add_nodata_subplots(self, figure):
        self.txt_plot = figure.add_subplot(111)

    def _add_subplots(self, figure):
        self.img_plot = figure.add_subplot(211)
        self.skyimg_plot = figure.add_subplot(212)

    def _data_plot(self):
        # get the right category file from our dictionary
        try:
         fov = self.frames[self.fov_cat_obj]
        except:
         fov = self.frames[self.fov_cat_sky]
    #    try: 
    #     fov = self.frames[self.fov_cat_sky]

        fov.readImage(1)
        mask = self.frames[self.mask_cat]
        mask.readImage()

        imgdisp = pipeline_display.ImageDisplay()
        imgdisp.setLabels('X', 'Y')
        imgdisp.setAspect('equal')
        tooltip_sky = paragraph("""\
        White light image of the cube.
        A greyscale is used for pixels 
        that will be used as sky, the fluxes
        of all other pixels are marked in 
        red.
        """)
        tooltip_obj = paragraph("""\
        Different representation of the same image as 
        above. The roles of the grey scale and red
        pixels are reversed:
        Sky pixels are marked in red.  A greyscale is 
        used for all other pixels.
        """)
        imgdisp.display(self.img_plot, "Sky", tooltip_sky, fov.image, bpmimage=mask.image)
        imgdisp.display(self.skyimg_plot, "Object", tooltip_obj, fov.image, bpmimage=(1-mask.image))

    def _nodata_plot(self):
        # could be moved to reflex library?
        self.txt_plot.set_axis_off()
        text_nodata = "Data not found. Input files should contain these" \
                       " types:\n%s" % self.img_cat
        self.txt_plot.text(0.1, 0.6, text_nodata, color='#11557c',
                      fontsize=18, ha='left', va='center', alpha=1.0)
        self.txt_plot.tooltip = 'No data found'


    def increaseFloatNumber(self, point) :
        floatpoint = self.getCurrentParameterHelper('floatopt') + 1
        new_params = list()
        new_params.append(reflex.RecipeParameter(self.recipe_name,'floatopt',
                                                 value=str(floatpoint)))
        return new_params

    def setCurrentParameterHelper(self, helper) :
        self.getCurrentParameterHelper = helper

#This is the 'main' function
if __name__ == '__main__':
    from reflex_interactive_app import PipelineInteractiveApp

    # Create interactive application
    interactive_app = PipelineInteractiveApp(enable_init_sop=True)

    # get inputs from the command line
    interactive_app.parse_args()

    #Check if import failed or not
    if not import_success:
        interactive_app.setEnableGUI(False)

    #Open the interactive window if enabled
    if interactive_app.isGUIEnabled():
        #Get the specific functions for this window
        dataPlotManager = DataPlotterManager()

        #the dataPlotManager uses the in_sof_rec_orig for deciding which
        #parameters to offer for plotting:
        dataPlotManager.orig_files= interactive_app.inputs.in_sof_rec_orig.files

        interactive_app.setPlotManager(dataPlotManager)
        interactive_app.showGUI()
    else:
        interactive_app.set_continue_mode()

    #Print outputs. This is parsed by the Reflex python actor to
    #get the results. Do not remove
    interactive_app.print_outputs()
    
    sys.exit()
 



#!/usr/bin/env python
import reflex
import sys
import os
import_ok = 1
try:
    from astropy.io import fits as pyfits
except ImportError:
    try:
        import pyfits
    except ImportError:
        import_ok = 0
try:
    import math
except ImportError:
    import_ok = 0


if __name__ == '__main__':

    parser = reflex.ReflexIOParser()

    # Define inputs
    parser.add_option("-i", "--in_sof", dest="in_sof")
    # Define outputs
    parser.add_output("-o", "--out_sof", dest="out_sof")

    inputs = parser.get_inputs()
    outputs = parser.get_outputs()

    # Retrieve input
    in_sof = inputs.in_sof

    # Get the input files
    files = in_sof.files

    # Do the stuff
    if import_ok == 0:
        outputs.out_sof = inputs.in_sof
        parser.write_outputs()
        sys.exit()

    # Read the content of TRACE_TABLE_06 (static, if present)
    for file in files:
        if file.category == 'TRACE_TABLE_06':
            static = pyfits.open(file.name, mode='readonly')
            newdata = static[1].data

    # Read the content of the TRATE_TABLE-06.fits, produced by the pipeline
    for file in files:
        if os.path.basename(file.name) == 'TRACE_TABLE-06.fits':
            hdulist = pyfits.open(file.name, mode='update')
            DataTable = hdulist[1].data
            Width = DataTable.field('Width')
            total = sum(Width)
            # If there is a NaN, it replaces the content of TRATE_TABLE-06.fits
            # with that of the static table.
            if math.isnan(total):
                hdulist[1].data = newdata
                hdulist.flush()
                DataTable = hdulist[1].data
                Width = DataTable.field('Width')
                total = sum(Width)

    outputs.out_sof = inputs.in_sof
    parser.write_outputs()

    sys.exit()



# ========================================================================================================
#
# Lodovico Coccato & John Pritchard
#
# A python script to check that the specified python modules are available
# Two types of modules can be specified:
#
#   processing_modules: requried for correct execution of the workflow
#                       if any are missing the workflow is stopped.
#
#   interactive_modules: required for certain interactive features, but
#                        the workflow can run correctly without them
#
# ========================================================================================================


import reflex 
import sys
import os
import json
from optparse import OptionParser
import time
from importlib import import_module
message_r='The following python packages needed for reduction are missing:\n'
message_i='The following python packages needed for interactivity are missing:\n'
import_r=False
import_i=False

def findOneoOf(module_list) :
    foundOne=False
    for om in m.split("|") :
        try :
            cm=om.split(":")
            print(("importing %s" %(cm[0])))
            mod=import_module(cm[0])
            if len(cm) > 1  : print(("checking for klass %s" %(cm[1])))
            klass=getattr(mod, cm[1]) if len(cm) > 1 else None
            foundOne=True
            print(('found %s' %(om)))
        except :
            print(('FAILED to find %s' %(om)))
            pass
    return foundOne

if __name__ == '__main__':

#   MAIN MODULE ---

    parser = reflex.ReflexIOParser()
    parser.add_option("-i", "--interactive_mode", dest="interactive_mode", type='string')
    parser.add_option("-R", "--processing_modules", dest="processing_modules", type='string')
    parser.add_option("-I", "--interactive_modules", dest="interactive_modules", type='string')
    parser.add_output("-m", "--message", dest="message")
    parser.add_output("-t", "--trigger", dest="trigger")
    inputs  = parser.get_inputs()
    outputs = parser.get_outputs()

    ## Python Modules required for processing...
    if not inputs.processing_modules == '' :
        for m in inputs.processing_modules.split(',') :
            try:
                if not findOneoOf(m) : raise ImportError
            except ImportError :
                import_r=True
                message_r=message_r+' %s\n' %(m)
            except Exception as e:
                raise e

    ## Python Modules required for interactivity but not for processing...
    if not inputs.interactive_modules == '' :
        for m in inputs.interactive_modules.split(',') :
            try:
                if not findOneoOf(m) : raise ImportError
            except ImportError :
                import_i=True
                message_i=message_i+' %s\n' %(m)
            except Exception as e:
                raise e

    message='' 
    if import_i :
        message = message + message_i+(' \n')

    if import_r :
        message = message + message_r+(' \n')
        message = message+('The workflow will stop')
    
    if not import_r and import_i :
        message = message+('The workflow will continue but interactivity features will not work')

    if not import_r :
        outputs.trigger = True

    #broadcast the output message only if a module is missing.
    if import_i or import_r : 
        outputs.message=message

    if inputs.interactive_mode.lower() == 'false' and not message == '' :
        print(message)


    parser.write_outputs()
    sys.exit()
 


      
